/****************************************************************************
 * <Novell-copyright>
 * Copyright (c) 2001 Novell, Inc. All Rights Reserved.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public License
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, contact Novell, Inc.
 * 
 * To contact Novell about this file by physical or electronic mail, you 
 * may find current contact information at www.novell.com.
 * </Novell-copyright>
 ****************************************************************************/

#include <config.h>

#include <xpl.h>
#include <memmgr.h>
#include <logger.h>
#include <hulautil.h>
#include <hulaagent.h>
#include <mdb.h>
#include <nmap.h>
#include <nmlib.h>
#include <msgapi.h>
#include <connio.h>

#include "generic.h"

GAgentGlobals GAgent = {{0,}, 0, };

static void 
GAgentClientFree(void *clientp)
{
    GAgentClient *client = clientp;

    if (client->envelope) {
        MemFree(client->envelope);
    }

    MemPrivatePoolReturnEntry(client);
}

static int 
ProcessEntry(void *clientp,
             Connection *conn)
{
    int ccode;
    int envelopeLength;
    char *ptr;
    char qID[16];
    GAgentClient *client = clientp;

    client->conn = conn;

    ccode = HulaQueueAgentHandshake(client->conn, 
                                    client->line, 
                                    qID, 
                                    &envelopeLength);

    sprintf(client->line, "GAgent: %s", qID);
    XplRenameThread(XplGetThreadID(), client->line);

    if (ccode == -1) {
        return -1;
    }

    client->envelope = HulaQueueAgentReadEnvelope(client->conn, 
                                                  client->line, 
                                                  envelopeLength);

    if (client->envelope == NULL) {
        return -1;
    }
    
    /* Queue-agent specific code.  This code just prints out the envelope */
    ptr = client->envelope;
    while (*ptr) {
        switch(*ptr++) {
        case QUEUE_FROM :
            XplConsolePrintf("hulageneric: message from: %s\n", ptr);
            break;
        case QUEUE_CALENDAR_LOCAL:
            XplConsolePrintf("hulageneric: deliver to calendar: %s\n", ptr);
            break;
        case QUEUE_RECIP_LOCAL :
        case QUEUE_RECIP_MBOX_LOCAL :
            XplConsolePrintf("hulageneric: deliver to local: %s\n", ptr);
            break;
        case QUEUE_FLAGS:
            XplConsolePrintf("hulageneric: flags: %s\n", ptr);
            break;
        case QUEUE_ADDRESS:
            XplConsolePrintf("hulageneric: address: %s\n", ptr);
            break;
        case QUEUE_BOUNCE:
            XplConsolePrintf("hulageneric: bounce: %s\n", ptr);
            break;
        case QUEUE_DATE:
            XplConsolePrintf("hulageneric: date: %s\n", ptr);
            break;
        case QUEUE_ID:
            XplConsolePrintf("hulageneric: queue id: %s\n", ptr);
            break;
        case QUEUE_RECIP_REMOTE:
            XplConsolePrintf("hulageneric: deliver to remote: %s\n", ptr);
            break;
        case QUEUE_THIRD_PARTY:
            XplConsolePrintf("hulageneric: third party: %s\n", ptr);
            break;
        }
        HULA_ENVELOPE_NEXT(ptr);
    }

    /* The client struct, connection, and QDONE will be cleaned up by
     * the caller */
    return 0;
}

static void 
GenericAgentServer(void *ignored)
{
    XplRenameThread(XplGetThreadID(), AGENT_DN " Server");

    /* Listen for incoming queue items.  Call ProcessEntry with a
     * GAgentClient allocated for each incoming queue entry. */
    HulaQueueAgentListenWithClientPool(&GAgent.agent,
                                       GAgent.nmapConn,
                                       GAgent.threadPool,
                                       sizeof(GAgentClient),
                                       GAgentClientFree, 
                                       ProcessEntry,
                                       &GAgent.clientPool);

    /* Shutting down */
    XplConsolePrintf(AGENT_NAME ": Shutting down.\r\n");

    if (GAgent.nmapConn) {
        ConnClose(GAgent.nmapConn, 1);
        GAgent.nmapConn = NULL;
    }

    GenericAgentManagementShutdown();

    HulaThreadPoolShutdown(GAgent.threadPool);
    HulaAgentShutdown(&GAgent.agent);

    XplConsolePrintf(AGENT_NAME ": Shutdown complete\r\n");
}

static BOOL 
ReadConfiguration(void)
{
    MDBValueStruct *config;

    config = MDBCreateValueStruct(GAgent.agent.directoryHandle, 
                                  MsgGetServerDN(NULL));
    if (config) {
        /* Read your agent's configuration here */
    } else {
        return FALSE;
    }

    MDBDestroyValueStruct(config);

    return TRUE;
}

static void 
SignalHandler(int sigtype) 
{
    HulaAgentHandleSignal(&GAgent.agent, sigtype);
}

XplServiceCode(SignalHandler)

int
XplServiceMain(int argc, char *argv[])
{
    int ccode;
    int minThreads;
    int maxThreads;
    int minSleep;

    if (XplSetRealUser(MsgGetUnprivilegedUser()) < 0) {
        XplConsolePrintf(AGENT_NAME ": Could not drop to unprivileged user '%s'\r\n" AGENT_NAME ": exiting.\n", MsgGetUnprivilegedUser());
        return -1;
    }

    strcpy(GAgent.nmapAddress, "127.0.0.1");

    /* Initialize the Hula libraries */
    ccode = HulaAgentInit(&GAgent.agent, AGENT_NAME, AGENT_DN);
    if (ccode == -1) {
        XplConsolePrintf(AGENT_NAME ": Exiting.\r\n");
        return -1;
    }

    /* Set up socket for listening on an incoming queue */ 
    GAgent.queueNumber = Q_FIVE;
    GAgent.nmapConn = HulaQueueConnectionInit(&GAgent.agent, GAgent.queueNumber);

    if (!GAgent.nmapConn) {
        HulaAgentShutdown(&GAgent.agent);
        XplConsolePrintf(AGENT_NAME ": Exiting.\r\n");
        return -1;
    }

    HulaQueueAgentGetThreadPoolParameters(&GAgent.agent,
                                          &minThreads, &maxThreads, &minSleep);
    
    /* Create a thread pool for managing connections */
    GAgent.threadPool = HulaThreadPoolNew(AGENT_DN " Clients", 
                                          minThreads, maxThreads, minSleep);

    if (GAgent.threadPool == NULL) {
        HulaAgentShutdown(&GAgent.agent);
        XplConsolePrintf(AGENT_NAME ": Unable to create thread pool.\r\n" AGENT_NAME ": Exiting.\r\n");
        return -1;
    }
    
    ReadConfiguration();

    XplSignalHandler(SignalHandler);

    GenericAgentManagementStart();

    /* Start the server thread */
    XplStartMainThread(AGENT_NAME, &id, GenericAgentServer, 8192, NULL, ccode);
    
    XplUnloadApp(XplGetThreadID());
    
    return 0;
}
