/****************************************************************************
 * DONE:
 * Add user
 * Delete user (does not remove their mailbox though)
 * Delete mailing list
 * Delete mailing list member
 *
 * TODO:
 * modify users (i.e. change attributes, etc)
 * Verify all return codes
 * x.500 to MDB name conversion
 *
 * Farther down the road...
 * add rest of Mailing list management (add/modify)
 ****************************************************************************/
#include <config.h>
#include "hulaadmin.h"
#include <xpl.h>
#include <memmgr.h>
#include <hulautil.h>
#include <mwtempl.h>
#include <webadmin.h>
#include <errno.h>
#include <connio.h>

enum HulaAdminActions {
    HULA_ADMIN_ADD,
    HULA_ADMIN_MODIFY,
    HULA_ADMIN_DELETE,
};

enum HulaAdminCommand {
    HULA_ADMIN_USER,
    HULA_ADMIN_LIST,
    HULA_ADMIN_LIST_USER,
};

struct _HULA_GLOBAL_VARIABLES {
    MDBHandle directoryHandle;

    struct {
        unsigned char name[MDB_MAX_OBJECT_CHARS + 1];
        unsigned char password[MDB_MAX_OBJECT_CHARS + 1];
        unsigned char context[MDB_MAX_OBJECT_CHARS + 1];
        unsigned char givenname[MDB_MAX_OBJECT_CHARS + 1];
        unsigned char surname[MDB_MAX_OBJECT_CHARS + 1];
    } user;
    
    struct {
        unsigned char name[MDB_MAX_OBJECT_CHARS + 1];
        unsigned char context[MDB_MAX_OBJECT_CHARS + 1];
    } list;
    
    enum HulaAdminActions action;
    enum HulaAdminCommand command;

    struct {
        unsigned char x500Name[MDB_MAX_OBJECT_CHARS + 1];
        unsigned char name[MDB_MAX_OBJECT_CHARS + 1];
        unsigned char password[MDB_MAX_OBJECT_CHARS + 1];
    } authentication;      
    BOOL verbose;
} HulaAdmin;

BOOL CheckRequired(void);

static void 
HulaAdminUsage(void)
{
    XplConsolePrintf("%s", NL);
    XplConsolePrintf("hulaadmin - Hula Admin Utility.%s", NL);
    XplConsolePrintf("usage: hulaadmin <command> <options>%s", NL);
    XplConsolePrintf("Type \"hulaadmin help <command>\" for help with a specific command.%s%s", NL, NL);
    XplConsolePrintf("Supported commands are:%s", NL);
    XplConsolePrintf("  user                        Add/Delete/Modify a User.%s", NL);
    XplConsolePrintf("  list                        Add/Delete/Modify a Mailing List.%s", NL);
    XplConsolePrintf("  list-user                   Add/Delete/Modify a Mailing List member.%s", NL);
    XplConsolePrintf("  help                        Help%s", NL);
    XplConsolePrintf("  --verbose                   Display verbose progress messages.%s", NL);

    return;
}
static void 
HulaAdminListUsage(void)
{
    XplConsolePrintf("%s", NL);
    XplConsolePrintf("hulaadmin - Hula Admin Utility.%s", NL);
    XplConsolePrintf("usage: hulaadmin list delete -u=<> -p=<> -L=<>%s", NL);
    XplConsolePrintf("usage: hulaadmin list delete-member -u=<> -p=<> -U=<> [ -L=<> | --fromall ]%s%s", NL, NL);
    XplConsolePrintf("  -U=<username>               The member to remove from the list.%s", NL);
    XplConsolePrintf("  -L=<listname>               The list to remove.%s", NL);
    XplConsolePrintf("  --from-all                  Remove the member from all lists (not implemented yet).%s", NL);
    XplConsolePrintf("  -p= | --pass=<text>         The password used for MDB authentication.%s", NL);
    XplConsolePrintf("  -u= | --user=<name>         A user name for MDB authentication.%s", NL);
    XplConsolePrintf("                                example: \\\\Tree\\\\Context\\\\Admin%s", NL);
    XplConsolePrintf("  --verbose                   Display verbose progress messages.%s", NL);
    return;
}

static void 
HulaAdminListUserUsage(void)
{
    XplConsolePrintf("%s", NL);
    XplConsolePrintf("hulaadmin - Hula Admin Utility.%s", NL);
    XplConsolePrintf("usage: hulaadmin list-user delete -u=<> -p=<> -U=<> [ -L=<> | --fromall ]%s", NL);
    XplConsolePrintf("  -U=<username>               The member to remove from the list.%s%s", NL, NL);
    XplConsolePrintf("  -L=<listname>               The list to remove.%s", NL);
    XplConsolePrintf("  --from-all                  Remove the member from all lists (not implemented yet).%s", NL);
    XplConsolePrintf("  -p= | --pass=<text>         The password used for MDB authentication.%s", NL);
    XplConsolePrintf("  -u= | --user=<name>         A user name for MDB authentication.%s", NL);
    XplConsolePrintf("                                example: \\\\Tree\\\\Context\\\\Admin%s", NL);
    XplConsolePrintf("  --verbose                   Display verbose progress messages.%s", NL);
    return;
}

static void 
HulaAdminUserUsage(void)
{
    XplConsolePrintf("%s", NL);
    XplConsolePrintf("hulaadmin - Hula Admin Utility.%s", NL);
    XplConsolePrintf("usage: hulaadmin user add -u=<> -p=<> -U=<> -P=<> -c=<> -s=<> [-g=<>]%s", NL);
    XplConsolePrintf("usage: hulaadmin user delete -u=<> -p=<> -U=<> -c=<>%s%s", NL, NL);
    XplConsolePrintf("  -U=<user>                   The username to use for the new user.%s", NL);
    XplConsolePrintf("  -P=<text>                   The password to use for the new user.%s", NL);
    XplConsolePrintf("  -c= | --context=<text>      The context where the user should be created.%s", NL);
    XplConsolePrintf("  -g= | --givenname=<text>    The Given Name of the new user.%s", NL);
    XplConsolePrintf("  -s= | --surname=<text>      The Surname of the new user.%s", NL);
    XplConsolePrintf("  -p= | --pass=<text>         The password used for MDB authentication.%s", NL);
    XplConsolePrintf("  -u= | --user=<name>         A user name for MDB authentication.%s", NL);
    XplConsolePrintf("                                example: \\\\Tree\\\\Context\\\\Admin%s", NL);
    XplConsolePrintf("  --verbose                   Display verbose progress messages.%s", NL);
	XplConsolePrintf("Examples:%s",NL);
	XplConsolePrintf("hulaadmin user add -u=\\\\Tree\\\\Context\\\\Admin -p=hula -U=scott -P=villinski -s=Villinski -c=Context%s",NL);
	XplConsolePrintf("hulaadmin user delete -u=\\\\Tree\\\\Context\\\\Admin -p=hula -U=scott -c=Context%s",NL);
    return;
}
BOOL
CheckRequired (void)
{
    /* this is where we check to make sure that the required parametes have been entered */
    BOOL passedCheck = TRUE;
    
    switch (HulaAdmin.command) {        
        case HULA_ADMIN_USER: {
            /* User actions */
            switch (HulaAdmin.action) {
                case HULA_ADMIN_ADD: {
                    /* user add requires: context, username, password, surname */
                    if (!HulaAdmin.user.name[0]) {
                        XplConsolePrintf("You must specify a username for the new user.%s", NL);
                        passedCheck = FALSE;
                    }
                    if (!HulaAdmin.user.password[0]) {
                        XplConsolePrintf("You must specify a password for the new user.%s", NL);
                        passedCheck = FALSE;
                    }
                    if (!HulaAdmin.user.context[0]) {
                        XplConsolePrintf("You must specify a context for the new user.%s", NL);
                        passedCheck = FALSE;
                    }
					if (!HulaAdmin.user.surname[0]) {
						XplConsolePrintf("You must specify a surname for the new user.%s",NL);
						passedCheck = FALSE;
					}
                    break;
                } case HULA_ADMIN_DELETE: {
                    /* user delete requires: context, username */
                    if (!HulaAdmin.user.name[0]) {
                        XplConsolePrintf("You must specify the username to delete.%s", NL);
                        passedCheck = FALSE;
                    }
                    if (!HulaAdmin.user.context[0]) {
                        XplConsolePrintf("You must specify the context for the user.%s", NL);
                        passedCheck = FALSE;
                    }
                    break;
                } default :{
                    passedCheck = FALSE;
                    break;
                }
            }
            break;
        } case HULA_ADMIN_LIST: {
            /* List actions */
            switch (HulaAdmin.action) {
                case HULA_ADMIN_DELETE: {
                    /* List delete requires mailing list */
                    if (!HulaAdmin.list.name[0]) {
                        XplConsolePrintf("You must specify a Mailing List to delete.%s", NL);
                        passedCheck = FALSE;
                    }
                    break;
                } default : {
                    passedCheck = FALSE;
                    break;                
                }
            }
            break;
        } case HULA_ADMIN_LIST_USER: {
            /* List-user actions */
            switch (HulaAdmin.action) {                
                case HULA_ADMIN_DELETE: {
                    /* List-user delete requires user name and mailing list name */
                    if (!HulaAdmin.list.name[0]){
                        XplConsolePrintf("You must specify a Mailing List or all lists.%s", NL);
                        passedCheck = FALSE;
                    }                
                    if (!HulaAdmin.user.name[0]) {
                        XplConsolePrintf("You must specify the member to remove.%s", NL);
                        passedCheck = FALSE;
                    }
                    break;
                } default : {
                    passedCheck = FALSE;
                    break;
                }
            }
            break;
        } default:
            HulaAdminUsage();
    }
    return (passedCheck);
}

static void
PrintUsage(void)
{
    switch (HulaAdmin.command) {
        /* Print out the usage for the various commands */
        case HULA_ADMIN_USER: {
            HulaAdminUserUsage();
            break;
        } case HULA_ADMIN_LIST: {
            HulaAdminListUsage();
            break;
        } case HULA_ADMIN_LIST_USER: {
            HulaAdminListUserUsage();
            break;
        } default: {
            HulaAdminUsage();
        }
    }
}

static void
CreateUser(MDBValueStruct *v, MDBValueStruct *data, MDBValueStruct *attribute)
{
    BOOL result = FALSE;
    if (HulaAdmin.verbose) {
        XplConsolePrintf("Checking to see if the context: %s already exists.%s", HulaAdmin.user.context, NL);
    }
    if (MDBIsObject(HulaAdmin.user.context, v) == TRUE) {
        MDBSetValueStructContext(HulaAdmin.user.context, v);
        attribute = MDBShareContext(v);
        data = MDBShareContext(v);
                
        if (HulaAdmin.verbose) {            
            XplConsolePrintf("Checking to see if user: %s already exists.%s", HulaAdmin.user.name, NL);
        }
        if (MDBIsObject(HulaAdmin.user.name, v) == FALSE) {
            /* Check to see if the user already exists */
            if (HulaAdmin.verbose) {
                XplConsolePrintf("Creating user: %s%s", HulaAdmin.user.name, NL);
            }
            /* Create the user */
            result = MDBCreateObject(HulaAdmin.user.name, "User", attribute, data, v);
            if (!result) {
                XplConsolePrintf("Failed to create user: %s%s", HulaAdmin.user.name, NL);
            } else {
                XplConsolePrintf("Successfully created new user: \\%s\\%s%s", HulaAdmin.user.context, HulaAdmin.user.name, NL);
            }
                        
            if (HulaAdmin.user.givenname[0]) {
                /* Add Given Name (if passed) */
                if (HulaAdmin.verbose) {
                    XplConsolePrintf("Adding Given Name: %s%s", HulaAdmin.user.givenname, NL);
                }
                result = MDBAdd(HulaAdmin.user.name, "Given Name", HulaAdmin.user.givenname, v);
                if (!result) {
                        XplConsolePrintf("Failed to set Given Name for user: %s%s", HulaAdmin.user.name, NL);
                }
            }
                        
            if (HulaAdmin.user.surname[0]) {
                /* Add Surname (if passed) */
                if (HulaAdmin.verbose) {
                    XplConsolePrintf("Adding Surname: %s%s", HulaAdmin.user.surname, NL);
                }
                result = MDBAdd(HulaAdmin.user.name, "Surname", HulaAdmin.user.surname, v);
                if (!result) {
                        XplConsolePrintf("Failed to set Surname for user: %s%s", HulaAdmin.user.name, NL);
                }
            }
                        
            if (HulaAdmin.verbose) {
                XplConsolePrintf("Setting password for: %s%s", HulaAdmin.user.name, NL);
            }
            /* Set the new user's password */
            result = MDBChangePasswordEx(HulaAdmin.user.name, NULL, HulaAdmin.user.password, v);
            if (!result) {
                    XplConsolePrintf("Failed to set password for user: %s%s", HulaAdmin.user.name, NL);
            }
        } else {
            XplConsolePrintf("The user: %s already exists in the context: %s.%s", HulaAdmin.user.name, HulaAdmin.user.context, NL);
        }
    } else {
        XplConsolePrintf("The context: %s does not exist, please create the context before proceeding.%s", HulaAdmin.user.context, NL);
    }
}

static void
DeleteUser (MDBValueStruct *v)
{
    BOOL result = FALSE;
    if (HulaAdmin.verbose) {
        XplConsolePrintf("Checking to see if the context: %s exists.%s", HulaAdmin.user.context, NL);
    }
    if (MDBIsObject(HulaAdmin.user.context, v) == TRUE) {
        MDBSetValueStructContext(HulaAdmin.user.context, v);
        
        if (HulaAdmin.verbose) {
            XplConsolePrintf("Checking to see if user: %s  exists.%s", HulaAdmin.user.name, NL);
        }
        if (MDBIsObject(HulaAdmin.user.name, v) == TRUE) {
            /* Check to see if the user exists */            
            if (HulaAdmin.verbose) {
                XplConsolePrintf("Deleting user: %s%s", HulaAdmin.user.name, NL);
            }
            /* Delete the user */
            result = MDBDeleteObject(HulaAdmin.user.name, FALSE, v);
            if (!result) {
                XplConsolePrintf("Failed to delete user: %s%s", HulaAdmin.user.name, NL);
            } else {
                XplConsolePrintf("Successfully deleted user: \\%s\\%s%s", HulaAdmin.user.context, HulaAdmin.user.name, NL);
            }
        } else {
            XplConsolePrintf("The user: %s does not exist in the context: %s.%s", HulaAdmin.user.name, HulaAdmin.user.context, NL);
        }
    } else {
        XplConsolePrintf("The context: %s does not exist.%s", HulaAdmin.user.context, NL);
    }
}


static void
DeleteList (MDBValueStruct *v)
{
    BOOL result = FALSE;    
    if (HulaAdmin.verbose) {
        XplConsolePrintf("Checking to see if the Mailing List container: %s exists.%s", HulaAdmin.list.context, NL);
    }
    if (MDBIsObject(HulaAdmin.list.context, v) == TRUE) {
        MDBSetValueStructContext(HulaAdmin.list.context, v);
        
        if (HulaAdmin.verbose) {
            XplConsolePrintf("Checking to see if the Mailing List: %s  exists.%s", HulaAdmin.list.name, NL);
        }
        if (MDBIsObject(HulaAdmin.list.name, v) == TRUE) {
            /* Check to see if the list exists */            
            if (HulaAdmin.verbose) {
                XplConsolePrintf("Deleting Mailing List: %s%s", HulaAdmin.list.name, NL);
            }
            /* Delete the list */
            /* Besure to set the Recurse Boolean to TRUE - to delete all sub objects */
            result = MDBDeleteObject(HulaAdmin.list.name, TRUE, v);
            if (!result) {
                XplConsolePrintf("Failed to delete Mailing List: %s%s", HulaAdmin.list.name, NL);
            } else {
                XplConsolePrintf("Successfully deleted Mailing List: \\%s%s%s", HulaAdmin.list.context, HulaAdmin.list.name, NL);
            }
        } else {
            XplConsolePrintf("The Mailing List: %s does not exist.%s", HulaAdmin.list.name, NL);
        }
    } else {
        XplConsolePrintf("The Mailing List container: %s does not exist.%s", HulaAdmin.list.context, NL);
    }
}

static void
DeleteListUser (MDBValueStruct *v)
{   
    BOOL result = FALSE;    
    if (HulaAdmin.verbose) {
        XplConsolePrintf("Checking to see if the Mailing List: %s exists.%s", HulaAdmin.list.name, NL);
    }
    if (MDBIsObject(strcat(HulaAdmin.list.context, HulaAdmin.list.name), v) == TRUE) {
        MDBSetValueStructContext(HulaAdmin.list.context, v);
        
        if (HulaAdmin.verbose) {
            XplConsolePrintf("Checking to see if the address: %s is a member of the Mailing List: %s  exists.%s", HulaAdmin.list.name, HulaAdmin.user.name, NL);
        }
        if (MDBIsObject(HulaAdmin.user.name, v) == TRUE) {
            /* Check to see if the user is a member of the Mailing List exists */            
            if (HulaAdmin.verbose) {
                XplConsolePrintf("Deleting Mailing List member: %s%s", HulaAdmin.user.name, NL);
            }
            /* Delete the user */
            result = MDBDeleteObject(HulaAdmin.user.name, FALSE, v);
            if (!result) {
                XplConsolePrintf("Failed to delete Mailing List member: %s%s", HulaAdmin.user.name, NL);
            } else {
                XplConsolePrintf("Successfully deleted member: %s from Mailing List: %s%s", HulaAdmin.user.name, HulaAdmin.list.name, NL);
            }
        } else {
            XplConsolePrintf("The address: %s is not a member of Mailing List: %s.%s", HulaAdmin.user.name, HulaAdmin.list.name, NL);
        }
    } else {
        XplConsolePrintf("The Mailing List: %s does not exist.%s", HulaAdmin.list.name, NL);
    }
}

int
main(int argc, char *argv[])
{
    int i = 0;
    BOOL result = TRUE;
    MDBValueStruct *v;
    MDBValueStruct *attribute = NULL;
    MDBValueStruct *data = NULL;
    
    /* initialize variables */
    HulaAdmin.command = -1;
    HulaAdmin.action = -1;
    HulaAdmin.user.name[0] = '\0';
    HulaAdmin.user.context[0] = '\0';
    HulaAdmin.user.password[0] = '\0';
    HulaAdmin.user.givenname[0] = '\0';
    HulaAdmin.user.surname[0] = '\0';
    HulaAdmin.list.name[0] = '\0';
    HulaAdmin.verbose = FALSE;
    strcpy(HulaAdmin.list.context, MAILING_LIST_CONTEXT);

    
    result = MemoryManagerOpen("Hula Admin");

    if (!result) {
        XplConsolePrintf("Failed to initialize the memory pool library.%s", NL);
        return(-1);
    }
    
    for (i = 1; i < argc; i++) {
        /* Read the command line arguments */        
        if ((XplStrCaseCmp(argv[i], "--help") == 0) 
                || (XplStrCaseCmp(argv[i], "help") == 0) 
                || (XplStrCaseCmp(argv[i], "-?") == 0)) {
                    /* Check to see if help was specified */
                    result = FALSE;
        
        } else if (XplStrCaseCmp(argv[i], "user") == 0) {
            HulaAdmin.command = HULA_ADMIN_USER;                            
        } else if (XplStrCaseCmp(argv[i], "list") == 0) {
            HulaAdmin.command = HULA_ADMIN_LIST;
        } else if (XplStrCaseCmp(argv[i], "list-user") == 0) {
            HulaAdmin.command = HULA_ADMIN_LIST_USER;
        } else if (XplStrCaseCmp(argv[i], "add") == 0) {
            HulaAdmin.action = HULA_ADMIN_ADD;
        } else if (XplStrCaseCmp(argv[i], "modify") == 0) {
            HulaAdmin.action = HULA_ADMIN_MODIFY;
        } else if (XplStrCaseCmp(argv[i], "delete") == 0) {
            HulaAdmin.action = HULA_ADMIN_DELETE;
        } else if (strncmp(argv[i], "-u=", 3) == 0) {
            /* Read the username to Authenticate to Hula */
            strcpy(HulaAdmin.authentication.name, argv[i] + 3);
        } else if (strncmp(argv[i], "--user=", 7) == 0) {
            /* Read the username to Authenticate to Hula */
            strcpy(HulaAdmin.authentication.name, argv[i] + 7);
                    
        } else if (strncmp(argv[i], "-p=", 3) == 0) {
            /* Read the password to authenticate to Hula */
            strcpy(HulaAdmin.authentication.password, argv[i] + 3);
        } else if (strncmp(argv[i], "--pass=", 7) == 0) {
            /* Read the password to authenticate to Hula */
            strcpy(HulaAdmin.authentication.password, argv[i] + 7);
                
        } else if (strncmp(argv[i], "--verbose", 9) == 0) {
            /* Check to see if verbose was enabled */
            HulaAdmin.verbose = TRUE;        
                
        } else if (strncmp(argv[i], "-U=", 3) == 0) {
            /* Read the username */
            strcpy(HulaAdmin.user.name, argv[i] + 3);
                    
        } else if (strncmp(argv[i], "-P=", 3) == 0) {
            /* read the password for the new user */
            strcpy(HulaAdmin.user.password, argv[i] + 3);
                    
        } else if (strncmp(argv[i], "-g=", 3) == 0) {
            /* read the Given Name for the new user */
            strcpy(HulaAdmin.user.givenname, argv[i] + 3);
        } else if (strncmp(argv[i], "--givenname=", 12) == 0) {
            /* read the Given Name for the new user */
            strcpy(HulaAdmin.user.givenname, argv[i] + 12);
                    
        } else if (strncmp(argv[i], "-s=", 3) == 0) {
            /* Read the Surname for the new user */
            strcpy(HulaAdmin.user.surname, argv[i] + 3);
        } else if (strncmp(argv[i], "--surname=", 10) == 0) {
            /* Read the Surname for the new user */
            strcpy(HulaAdmin.user.surname, argv[i] + 10);
        
        } else if (strncmp(argv[i], "-c=", 3) == 0) {
            /* Read the Context */
            strcpy(HulaAdmin.user.context, argv[i] + 3);
        } else if (strncmp(argv[i], "--context=", 11) == 0) {
            /* Read the Context */
            strcpy(HulaAdmin.user.context, argv[i] + 11);
                
        } else if (strncmp(argv[i], "-L=", 3) == 0) {
            /* Read the listname */
            strcpy(HulaAdmin.list.name, argv[i] + 3);
        }
    }
        
    if ((!result) || (argc <= 2)) {
        /* Print out the Hula Admin help if no arguments or --help was specified */
        PrintUsage();
        return(0);
    }

    if (!CheckRequired()) {
        /* Check to make certain that we have the required data based on the action to perform */
        /* If we fail - print out the usage for the specified command */
        PrintUsage();
        return(0);        
    }
    
    /* If we get there - that means thins are looking up! */
        
    if (HulaAdmin.verbose) {
        XplConsolePrintf("Initializing MDB.%s", NL);
    }
    if (!MDBInit()) {
        /* Init MDB */
        XplConsolePrintf("hulaadmin: Exiting\n.");
        return 1;
    }

    if (HulaAdmin.verbose) {
        XplConsolePrintf("Authenticating to MDB.%s", NL);
    }
    if (HulaAdmin.authentication.name[0]) {
        /* Authenticate to MDB */
        if (HulaAdmin.verbose) {
            XplConsolePrintf("Authenticating as: %s%s", HulaAdmin.authentication.name, NL);
        }
        HulaAdmin.directoryHandle = MDBAuthenticate("Hula", HulaAdmin.authentication.name, HulaAdmin.authentication.password);
    } else {
        /* if no username is provided, we'll auth as public user - mileage may very */
        if (HulaAdmin.verbose) {
            XplConsolePrintf("Performing public authentication.%s", NL);
        }
        HulaAdmin.directoryHandle = MDBAuthenticate("Hula", NULL, NULL);
    }
        
    if (HulaAdmin.directoryHandle) {
        /* Verify that we have a directoryHandle to talk to MDB */
        /* get the context in MDB */
        v = MDBCreateValueStruct(HulaAdmin.directoryHandle, NULL);  
        switch (HulaAdmin.command) {
            case HULA_ADMIN_USER: {
                switch (HulaAdmin.action) {
                    case HULA_ADMIN_ADD: {
                        /* call CreateUser() */
                        CreateUser(v, data, attribute);
                        break;
                    } case HULA_ADMIN_DELETE: {
                        /* call DeleteUser() */
                        DeleteUser(v);
                        break;
                    } default: {
                        break;
                    }
                }
                break;
            } case HULA_ADMIN_LIST: {
                switch (HulaAdmin.action) {
                    case HULA_ADMIN_DELETE: {
                        /* call DeleteList() */
                        DeleteList(v);
                        break;
                    } default: {
                        break;
                    }
                }
                break;
            } case HULA_ADMIN_LIST_USER: {
                switch (HulaAdmin.action) {
                    case HULA_ADMIN_DELETE: {
                        /* call DeleteListUser() */
                        DeleteListUser(v);
                        break;
                    } default: {
                        break;
                    }
                }
                break;
            } default: {
                break;
            }
        }
    }
    
    if (HulaAdmin.verbose) {
        XplConsolePrintf("Logging out of MDB.%s", NL);
    }

    if (!MDBRelease(HulaAdmin.directoryHandle)) {
        XplConsolePrintf("Failed to release MDBHandle.%s", NL);
    }

    MemoryManagerClose("Hula Admin");
    
    return 0;
}
