using System;
using System.Xml;
using System.Web;
using System.Text;
using System.Collections;
using System.IO;
using Lucene.Net.Search;
using Lucene.Net.Index;
using Lucene.Net.Documents;
using Hula.Store;
using Hula.Util;
using Hula.Cal;
using Hula.ICal;

namespace Hula.Dav {

public class CalQueryReport : QueryReport {
    public CalQueryReport(HttpContext context) : base (context) {
    }

    protected override void SetupQuery() {
	base.SetupQuery();
	query.Add(new TermQuery(new Term("hula.mime-type", "text/calendar")), true, false);
    }

    protected override void HandleNode(XmlNode node) {
	if (node.LocalName == "filter") {
	    ParseFilter(node);
	} else {
	    base.HandleNode(node);
	}
    }

    private void ParseComponentTimeRange(XmlNode node) {
	XmlAttribute startAttr = node.Attributes["start"];
	XmlAttribute endAttr = node.Attributes["end"];
	
	/* objects where dtstart is before the *end* time */
	if (startAttr != null) {
	    ICalTime start = ICalTime.Parse(startAttr.Value);
	    RangeFilter startFilter = RangeFilter.More("hula.end-time", start.ToString(DateTools.Resolution.MINUTE));
	    AddLuceneFilter(startFilter, ChainedFilter.Operation.And);
	}
	
	/* and dtend is after the *start* time */
	if (endAttr != null) {
	    ICalTime end = ICalTime.Parse(endAttr.Value);
	    RangeFilter endFilter = RangeFilter.Less("hula.start-time", end.ToString(DateTools.Resolution.MINUTE));
	    AddLuceneFilter(endFilter, ChainedFilter.Operation.And);
	}
    }

    private void ParseComponentParamFilter (XmlNode node, string prop) {
	string name = node.Attributes["name"].Value;
	
	foreach (XmlNode child in node) {
	    switch (child.LocalName) {
	    case "is-defined" :		
		break;
	    case "text-match" :
		// FIXME: Don't do caseless param filters for now.
		TermQuery q = new TermQuery (new Term("ical." + prop + ".param", + name + "=" + child.InnerText));

		query.Add(q, true, false);
		break;
	    }
	}
    }

    private void ParseComponentPropFilter (XmlNode node) {
	string name = node.Attributes["name"].Value.ToLower();
	
	foreach (XmlNode child in node) {
	    switch (child.LocalName) {
	    case "is-defined" :		
		break;
	    case "time-range" :
		break;
	    case "text-match" :
		XmlAttribute caselessAttr = child.Attributes["caseless"];
		string value;
		TermQuery q;
		if (caselessAttr != null && caselessAttr.Value == "yes") {
		    value = child.InnerText.ToLower();
		    q = new TermQuery (new Term("ical." + name + ".value.lower", value));
		} else {
		    q = new TermQuery (new Term("ical." + name + ".value", child.InnerText));
		}

		query.Add(q, true, false);
		break;
	    case "param-filter" :
		ParseComponentParamFilter(child, name);
		break;
	    }
	}	
    }

    private void ParseComponentFilter (XmlNode node) {
	foreach (XmlNode child in node) {
	    switch (child.LocalName) {
	    case "comp-filter" :
		break;
	    case "is-defined" :
		// implied
		break;
	    case "time-range" :
		ParseComponentTimeRange(child);
		break;
	    case "prop-filter" :
		ParseComponentPropFilter(child);
		break;
	    }
	}	
    }

    private void ParseCalendarFilter (XmlNode node) {
	foreach (XmlNode child in node) {
	    switch (child.LocalName) {
	    case "comp-filter" :
		string name = child.Attributes["name"].Value;
		TermQuery q = new TermQuery(new Term("ical.type", name));
		query.Add(q, true, false);

		ParseComponentFilter (child);
		break;
	    case "is-defined" :
		// implied 
		break;
	    case "time-range" :
		break;
	    case "prop-filter" :
		break;
	    }
	}
    }

    private void ParseFilter (XmlNode node) {
	foreach (XmlNode child in node) {
	    switch(child.LocalName) {
	    case "comp-filter" :
		string name = child.Attributes["name"].Value;
		if (name == "VCALENDAR") {
		    ParseCalendarFilter(child);
		}
		break;
	    case "is-defined" :
		// implied
		break;
	    }
	}
    }
}
}
