using System;
using System.Collections;
using Lucene.Net;
using Lucene.Net.Analysis;
using Lucene.Net.Analysis.Standard;
using Lucene.Net.Index;
using Lucene.Net.Search;
using Lucene.Net.Store;
using Lucene.Net.QueryParsers;

using log4net;

namespace Hula.Indexing
{
public class LuceneIndex : IDisposable
{
    private static Hashtable indexes = new Hashtable();

    private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

    private Directory dir = null;
    private Analyzer analyzer = null;
    private QueryParser parser  = null;

    private IndexReader reader = null;
    private IndexWriter writer = null;
    private IndexSearcher searcher = null;

    private string factoryKey = null;
    private int refcount = 1;
    
    public static LuceneIndex Get(string dir, Analyzer analyzer) {
	lock (indexes) {
	    LuceneIndex idx;
	    if (indexes.Contains(dir)) {
		idx = (LuceneIndex)indexes[dir];
		idx.refcount++;
	    } else {
		idx = new LuceneIndex(dir, analyzer);
		indexes[dir] = idx;
	    }
	    idx.factoryKey = dir;
	    return idx;
	}
    }

    public LuceneIndex(Directory dir,
		       Analyzer analyzer)
    {
        this.dir = dir;
        this.analyzer = analyzer;
    }

    private LuceneIndex(string path,
			Analyzer analyzer) : this(GetDirectory(path),
						  analyzer)
    {
    }

    private static Directory GetDirectory(string path)
    {
        if (System.IO.File.Exists(path + System.IO.Path.DirectorySeparatorChar + "segments"))
        {
            log.DebugFormat("using an existing Lucene directory: {0}", path);
            return FSDirectory.GetDirectory(path, false);
        }

        log.DebugFormat("creating a new Lucene index: {0}", path);
        System.IO.Directory.CreateDirectory(path);
        FSDirectory ret = FSDirectory.GetDirectory(path, true);
        new IndexWriter(ret, null, true).Close();
        return ret;
    }

    public void CloseReader()
    {
        lock(indexes) {
            if (reader != null)
            {
                log.DebugFormat("closing IndexReader: {0}", dir);
                reader.Close();
                reader = null;
            }
        }
    }

    public void Close()
    {
	lock(indexes) {
	    refcount--;
	    if (refcount != 0) {
		return;
	    }

            CloseReader();

	    if (writer != null)
	    {
		log.DebugFormat("optimizing index: {0}", dir);
		writer.Optimize();
		
		log.DebugFormat("closing IndexWriter: {0}", dir);
		writer.Close();
		writer = null;
	    }
	    if (searcher != null)
	    {
		log.DebugFormat("closing IndexSearcher: {0}", dir);
		searcher.Close();
		searcher = null;
	    }
	    if (factoryKey != null) {
		indexes.Remove(factoryKey);
	    }
	}
    }

    public Analyzer Analyzer
    {
        get
        {
            return analyzer;
        }
    }

    public QueryParser QueryParser
    {
        get
        {
	    if (parser == null) {
		parser = new QueryParser ("everything", analyzer);

	    }
            return parser;
        }
    }

    public IndexReader Reader
    {
        get
        {
            if (reader == null)
            {
                reader = IndexReader.Open(dir);
            }

            return reader;
        }
    }

    public IndexWriter Writer
    {
        get
        {
            if (writer == null)
            {
		writer = new IndexWriter(dir, analyzer, false);

                // optimize for indexing tons of documents
                writer.mergeFactor = 100;
                writer.maxMergeDocs = 1000;
	    }
	    
            return writer;
        }
    }

    public IndexSearcher Searcher
    {
        get
        {
            if (searcher == null)
            {
                searcher = new IndexSearcher(Reader);
            }

            return searcher;
        }
    }

    public void Dispose() {
	Close();
	GC.SuppressFinalize(this);
    }

    ~LuceneIndex() {
	Close();
    }

}
}
