/****************************************************************************
 *
 * Copyright (c) 2005 Novell, Inc.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail,
 * you may find current contact information at www.novell.com
 *
 ****************************************************************************/

using System;
using System.Collections;
using System.IO;
using System.Text;

using log4net;

namespace Hula.Nmap
{
public struct DocumentInfo {
    public string guid;
    public DocumentType type;
    public string resource;
    public int indexver;
    public string modified;
};

public enum DocumentTypeMasks {
    Unknown      = 0x0001,
    Mail         = 0x0002,
    Calendar     = 0x0003,
    Addressbook  = 0x0004,

    Folder       = 0x1000,
    Shared       = 0x2000,
    SharedFolder = Folder | Shared,
    Purged       = 0x4000,
}

public enum DocumentType {
    Unknown     = DocumentTypeMasks.Unknown,
    Mail        = DocumentTypeMasks.Mail,
    Calendar    = DocumentTypeMasks.Calendar,
    Addressbook = DocumentTypeMasks.Addressbook,
    
    UnknownFolder     = DocumentTypeMasks.Folder | Unknown,
    MailFolder        = DocumentTypeMasks.Folder | Mail,
    CalendarFolder    = DocumentTypeMasks.Folder | Calendar,
    AddressbookFolder = DocumentTypeMasks.Folder | Addressbook,

    UnknownSharedFolder     = DocumentTypeMasks.SharedFolder | Unknown,
    MailSharedFolder        = DocumentTypeMasks.SharedFolder | Mail,
    CalendarSharedFolder    = DocumentTypeMasks.SharedFolder | Calendar,
    AddressbookSharedFolder = DocumentTypeMasks.SharedFolder | Addressbook,
}

public class DocumentCommands : NmapCommands
{
    private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

    public DocumentCommands(NmapConnection nconn) : base(nconn) {
    }

    private static string ApplyFilter (Hula.Util.Filter filter, TextReader reader) {
	StringWriter writer = new StringWriter();
	filter.Run (reader, writer);
	return writer.ToString();
    }

    public void DSTOR(string folder, DocumentType type, TextReader reader, Hula.Util.Filter filter, out string guid, out string etag)
    {
	string s = ApplyFilter (filter, reader);

	nconn.WriteRequest("DSTOR {0} {1} {2}", folder, (int)type, s.Length);
	NmapResponse response = nconn.ReadFullReply();
	if (response.code != 2002) {
	    guid = etag = null;
	    throw new NmapException(response);
	} 

	nconn.Writer.Write (s);
	nconn.Writer.Flush ();

	response = nconn.ReadFullReply();
	if (response.code != 1000) {
	    throw new NmapException(response);
	}

	string[] responseSplit = response.text.Split(' ');
	guid = responseSplit[0];
	etag = responseSplit[1];
    }

    public void DREPL(string guid, string etag, TextReader reader, Hula.Util.Filter filter, out string newEtag)
    {
	string s = ApplyFilter (filter, reader);

	if (etag == null) {
	    etag = "-";
	}
	nconn.WriteRequest("DREPL {0} {1} {2}", guid, etag, s.Length);
	NmapResponse response = nconn.ReadFullReply();
	if (response.code != 2002) {
	    throw new NmapException(response);
	}

	nconn.Writer.Write (s);
	nconn.Writer.Flush ();
	
	response = nconn.ReadFullReply();
	if (response.code != 1000) {
	    throw new NmapException(response);
	}

	string[] responseSplit = response.text.Split(' ');
	newEtag = responseSplit[1];
    }

    public void DLIST(string guid, out IDictionary properties, out NmapStreamReader reader)
    {
	reader = null;

        nconn.WriteRequest("DLIST {0}", guid);

	properties = new Hashtable();

	NmapResponse response = nconn.ReadFullReply();

	while (response.code == 2001) {
	    string[] values = response.text.Split(' ');
	    properties[values[0]] = nconn.ReadBytes(Int32.Parse(values[1]));
	    response = nconn.ReadFullReply();
	}	

	if (response.code == 2002) {
	    string[] values = response.text.Split(' ');
	    reader = nconn.GetReader(Int32.Parse(values[0]));
	} else if (response.code == 2003) {
	    string[] values = response.text.Split(' ');
	    int numLines = Int32.Parse(values[0]);
	    for (int i = 0; i < numLines; i++) {
		nconn.ReadLine();
	    }
	    response = nconn.ReadFullReply();
	    reader = null;

	    if (response.code != 1000) {
		throw new NmapException(response);
	    }
	}

    }

    public IDictionary DPGET(string guid)
    {
        nconn.WriteRequest("DPGET {0}", guid);

	Hashtable properties = new Hashtable();

	NmapResponse response = nconn.ReadFullReply();
	while (response.code == 2001) {
	    string[] values = response.text.Split(' ');
	    properties[values[0]] = nconn.ReadBytes(Int32.Parse(values[1]));
	    response = nconn.ReadFullReply();	    
	}

	if (response.code != 1000) {
	    throw new NmapException(response);
	}

	return properties;
    }

    public void DPSET(string guid, string propname, string value) {
	nconn.WriteRequest("DPSET {0} {1} {2}", guid, propname, value.Length);
	NmapResponse response = nconn.ReadFullReply();
	if (response.code != 2002) {
	    throw new NmapException(response);
	} else {
	    nconn.Writer.Write(value);
	    nconn.Writer.Flush();
	}
	
	response = nconn.ReadFullReply();

	if (response.code != 1000) {
	    throw new NmapException(response);
	}
    } 

    public NmapStreamReader DREAD(string guid)
    {
        nconn.WriteRequest("DREAD {0}", guid);
	return nconn.GetReplyReader();
    }

    public void DPURG(string guid) {
	nconn.WriteRequest("DPURG {0}", guid);
	NmapResponse response = nconn.ReadFullReply();
	if (response.code != 1000) {
	    throw new NmapException(response);
	}
    }
    
    public void DCREA(DocumentType type, string path, out string guid, out string etag) {
	nconn.WriteRequest("DCREA {0} {1}", path, (int)type);
	NmapResponse response = nconn.ReadFullReply();
	if (response.code != 1000) {
	    throw new NmapException(response);
	}
	
	string[] responseSplit = response.text.Split(' ');
	guid = responseSplit[0];
	etag = responseSplit[1];
    }

    private DocumentInfo GetDocumentInfo(string info) 
    {
	DocumentInfo di = new DocumentInfo();

	string[] vals = info.Split(' ');

	di.guid = vals[0];
	di.resource = vals[1];
	di.type = (DocumentType)Int32.Parse(vals[2]);
	di.indexver = Int32.Parse(vals[3]);
	di.modified = vals[4];
	
	return di;
    }
    
    public IList DUNFO() {
	// FIXME: this function sucks
	nconn.WriteRequest("DUNFO");
        NmapResponse response = nconn.ReadLongReply ();

        if (response.data == null || response.data.Equals(""))
        {
            return new ArrayList();
        }

        string[] lines = response.data.Split('\n');
        IList ret = new ArrayList(lines.Length);

        for(int i=0; i<lines.Length; i++)
        {
            ret.Add(GetDocumentInfo(lines[i]));
        }

        return ret;
    }

    public IList DLSPG() {
	nconn.WriteRequest("DLSPG");
        NmapResponse response = nconn.ReadLongReply ();

        if (response.data == null || response.data.Equals(""))
        {
            return new ArrayList();
        }

        string[] lines = response.data.Split('\n');
        IList ret = new ArrayList(lines.Length);

        for(int i=0; i<lines.Length; i++)
        {
            ret.Add(lines[i]);
        }

        return ret;
    }

    public void DRMPG(IList guids) {
        StringBuilder args = new StringBuilder();

        foreach (string guid in guids) {
            args.Append(" ");
            args.Append(guid);
        }

	nconn.WriteRequest("DRMPG{0}", args.ToString());
        NmapResponse response = nconn.ReadLongReply ();
    }

    public IList DINFO(string[] guids) {
        nconn.WriteRequest("DINFO " + String.Join(" ", guids));
        NmapResponse response = nconn.ReadLongReply ();

        if (response.data == null || response.data.Equals("")) {
            return new ArrayList();
        }

        string[] lines = response.data.Split('\n');
        IList ret = new ArrayList(lines.Length);

        for(int i=0; i<lines.Length; i++) {
            if (lines[i].Equals("")) { // Document was not found
                continue;
            }

            ret.Add(GetDocumentInfo(lines[i]));
        }

        return ret;
    }

    public void DINSET(string guid, int indexNumber) {
	nconn.WriteRequest("DINSET {0} {1}", indexNumber, guid);
	NmapResponse response = nconn.ReadLongReply();
	if (response.code == 2000) {
	    response = nconn.ReadLongReply();
	    while (response.code == 2001 ||
		   response.code == 3011) {
		response = nconn.ReadFullReply();
	    }
	}

	if (response.code != 1000) {
	    throw new NmapException(response);
	}
    }

    public void DUNSET() {
	nconn.WriteRequest("DUNSET");
	NmapResponse response = nconn.ReadFullReply();

	if (response.code != 1000) {
	    throw new NmapException(response);
	}
    }
}
}
