using System;
using System.IO;
using System.Collections;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;

namespace Hula.Store {

//<summary>
// A simple store based on an on-disk directory filled with two files
// per object.  This store is intended for prototyping purposes only,
// so it isn't threadsafe/reliable.  Or even a little bit efficient.
//</summary>
public class DirectoryStore : FilteringStore, IStore, IDisposable {
    private string owner;
    private string storeDir;
    private string storeAlias;
    static private Hula.Util.Filter nullFilter = new Hula.Util.Filter();

    public string Owner { get { return owner; } }
    public string Name { get { return storeAlias; } }

    public DirectoryStore(string owner, string storeName) {
        this.owner = owner;
	this.storeAlias = storeName;
        storeDir = "store-" + owner;

	if (!Directory.Exists(storeDir)) {
	    Directory.CreateDirectory(storeDir);
	    if (storeAlias == "calendar" || storeAlias == "addressbook") {
		AddFolder("/Personal");
		AddFolder("/calendar");
	    }
	}
    }
    
    private string GetFilename(string guid) {
	return storeDir + "/" + guid;
    }
    
    private string GetPropertiesFilename(string guid) {
	return storeDir + "/" + guid + ".properties";
    }
    
    private void WriteFile(string guid, TextReader reader) {
	using (FileStream file = new FileStream(GetFilename(guid), FileMode.Create, FileAccess.Write, FileShare.None)) {
	    using (StreamWriter writer = new StreamWriter(file)) {
		nullFilter.Run(reader, writer);
	    }
	}
    }
    
    public void Add (string folder, 
		     IDictionary properties,
		     TextReader reader,
		     string mimeType,
		     out string guid,
		     out string etag) {
	
	guid = System.Guid.NewGuid().ToString();
	etag = "\"" + System.Guid.NewGuid().ToString() + "\"";

	if (!FolderExists(folder)) {
	    throw new DirectoryNotFoundException();
	}
	
	WriteFile(guid, reader);
	
	SetProperties(guid, properties);
	
	SetProperty(guid, "hula.mime-type", mimeType);
	SetProperty(guid, "hula.folder", folder);
	SetProperty(guid, "hula.etag", etag);
    }
    
    public TextReader Get (string guid) {
	if (guid == null) {
	    throw new ArgumentException("guid cannot be null");
	}

	return new StreamReader(File.OpenRead(GetFilename(guid)));
    }

    public void Get(string guid, out IDictionary properties, out TextReader reader) {
	reader = Get(guid);
	properties = GetProperties(guid);
    }
    
    private void CheckETag(string guid, string etag) {
	if (etag != null) {
	    string oldEtag = GetProperty(guid, "hula.etag");
	    if (oldEtag != etag) {
		throw new ETagException();
	    }
	}
    }

    public void Replace (string guid, string etag, TextReader reader, string mimeType, out string newEtag) {
	CheckETag(guid, etag);
	newEtag = "\"" + System.Guid.NewGuid().ToString() + "\"";
	
	SetProperty(guid, "hula.mime-type", mimeType);
	SetProperty(guid, "hula.etag", newEtag);
	WriteFile(guid, reader);
    }    

    public void Move (string guid, string etag, string newFolder) {
	CheckETag(guid, etag);
	SetProperty(guid, "hula.folder", newFolder);
    }
    
    public string Copy (string guid, string etag, string newFolder) {
	CheckETag(guid, etag);
	throw new NotImplementedException();
    }
    
    public void Delete (string guid, string etag) {
	CheckETag(guid, etag);
	if (FolderExists(guid)) {
	    DeleteFolder(guid);
	} else {
	    File.Delete(GetFilename(guid));
	}
	File.Delete(GetPropertiesFilename(guid));
    }
    
    public IDictionary GetProperties (string guid) {
	string filename = GetPropertiesFilename(guid);
	if (File.Exists(filename)) {
	    using (FileStream fs = File.OpenRead(filename)) {
		BinaryFormatter bf = new BinaryFormatter();
		return (IDictionary)bf.Deserialize(fs);
	    }
	} else {
	    return new Hashtable();
	}
    }
    
    public string GetProperty (string guid, string key) {
	IDictionary dict = GetProperties(guid);
	return (string)dict[key];
    }
    
    private void SetProperties (string guid, IDictionary dict) {
	using (FileStream fs = File.OpenWrite(GetPropertiesFilename(guid))) {
	    BinaryFormatter bf = new BinaryFormatter();
	    bf.Serialize(fs, dict);
	}
    }
    
    public void SetProperty (string guid, string key, string value) {
	IDictionary dict = GetProperties(guid);
	dict[key] = value;
	SetProperties(guid, dict);
    }

    private IDictionary GetFolders() {
	string filename = storeDir + "/" + storeAlias + ".directories";
	if (File.Exists(filename)) {
	    using (FileStream fs = File.OpenRead(filename)) {
		BinaryFormatter bf = new BinaryFormatter();
		return (IDictionary)bf.Deserialize(fs);
	    }
	} else {
	    return new Hashtable();
	}
    }

    private void SetFolders (IDictionary dict) {
	string filename = storeDir + "/" + storeAlias + ".directories";

	using (FileStream fs = File.OpenWrite(filename)) {
	    BinaryFormatter bf = new BinaryFormatter();
	    bf.Serialize(fs, dict);
	}
    }

    private void SplitPath(string path, out string folder, out string filename) {
        int i = path.IndexOf('/', 1);
        if (i > 0) {
            folder = path.Substring(0, i);
            filename = path.Substring(i + 1);
        } else {
            folder = "/";
            filename = path.Substring(1);
        }
    }

    public string AddFolder(string folder) {
        folder = folder.TrimEnd(new char[] {'/'});

	if (FolderExists(folder)) {
	    throw new ETagException();
	}

        string parent, child;
        SplitPath(folder, out parent, out child);

	if (parent != "/" && !FolderExists(parent)) {
	    throw new DirectoryNotFoundException();
	}

	if (storeAlias == "calendar" && 
            parent != "/") {
	    // FIXME: wrong exception
	    throw new ETagException();
	}

	IDictionary dict = GetFolders();
	dict[folder] = true;
	SetFolders(dict);

        SetProperty(folder, "hula.folder", parent);
        SetProperty(folder, "hula.filename", child);
        SetProperty(folder, "hula.mime-type", "application/x-directory");

        return folder;
    }

    private void DeleteFolder(string folder) {
        folder = folder.TrimEnd(new char[] {'/'});

	IDictionary dict = GetFolders();
	dict[folder] = true;
	SetFolders(dict);
    }

    private bool FolderExists(string folder) {
        folder = folder.TrimEnd(new char[] {'/'});

	IDictionary dict = GetFolders();
	return dict.Contains(folder);
    }

    public void Dispose() {
    }
}
}
